<?php
//
// $Id: tina_mvc_functions.php, v 0.00 Sat Jan 23 2010 21:47:32 GMT+0000 (IST) Francis Crossen $
//

/**
* General functions used throughout the Tina framework
*
* @package    Tina-MVC
* @subpackage Tina-Core-Helpers
* @author     Francis Crossen <francis@crossen.org>
*/

/**
 * Make a SQL datetime
 *
 * No checks are done on $ts parameter
 *
 * @param   integer $ts a unix timestamp. (Defaults to now)
 * @return  string  SQL datetime stamp ('YYYY-MM-DD HH:MM:SS')
 * @todo    Perhaps validation on the $ts parameter
 */
function tina_mvc_db_datetime($ts=FALSE) {
    
    if( $ts ) {
        return date( 'Y-m-d H:i:s', intval($ts) );
    }
    else {
        return date( 'Y-m-d H:i:s' );  
    }
    
}

/**
 * Format a SQL datetime
 *
 * No checks are done on $dt parameter
 *
 * @param   string $format a datetime format. See PHP date() function
 * @param   string $dt a MySQL datetime
 * @return  string  formatted datetime
 * @todo    Perhaps validation on the $dt parameter
 */
function tina_mvc_format_db_datetime( $format='', $dt=FALSE) {
    
    $dt = strptime( $dt, '%Y-%m-%d %H:%M:%S' );
    return date( $format, mktime( $dt['tm_hour'], $dt['tm_min'], $dt['tm_sec'], $dt['tm_mon'], $dt['tm_mday'], $dt['tm_year']+1900 ) );
    
}

/**
 * Make a SQL time
 *
 * No checks are done on $ts parameter
 *
 * @param   integer $ts a unix timestamp. (Defaults to now)
 * @return  string  SQL time stamp ('HH:MM:SS')
 * @todo    Perhaps validation on the $ts parameter
 */
function tina_mvc_db_time($ts=FALSE) {
    
    if( $ts ) {
        return date( 'H:i:s', intval($ts) );  
    }
    else {
        return date( 'H:i:s' );  
    }
    
}

/**
 * Make a SQL date
 *
 * No checks are done on $ts parameter
 *
 * @param   integer $ts a unix timestamp. (Defaults to now)
 * @return  string  SQL time stamp ('YYYY-MM-DD')
 * @todo    Perhaps validation on the $ts parameter
 */
function tina_mvc_db_date($ts=FALSE) {
    
    if( $ts ) {
        return date( 'Y-m-d', intval($ts) );  
    }
    else {
        return date( 'Y-m-d' );  
    }
    
}

/**
 * Get a value from $_POST
 *
 * A centralised place to get $_POST data
 *
 * Wordpress has some funky ways of treating global $_POST and $_GET variables.
 * Look at wp-settings.php (line 624 for v2.9.1):
 * <code>
 * // If already slashed, strip.
 * if ( get_magic_quotes_gpc() ) {
 *	$_GET    = stripslashes_deep($_GET   );
 *	$_POST   = stripslashes_deep($_POST  );
 *	$_COOKIE = stripslashes_deep($_COOKIE);
 * }
 *
 * // Escape with wpdb.
 * $_GET    = add_magic_quotes($_GET   );
 * $_POST   = add_magic_quotes($_POST  );
 * $_COOKIE = add_magic_quotes($_COOKIE);
 * $_SERVER = add_magic_quotes($_SERVER);
 * </code>
 * Tina assumes that you want to deal with unescaped data. If you want to store
 * it in a DB then do your own escaping
 *
 * @param   string  $var Variable name to retrieve
 * @return  mixed The $_POST var
 */
function tina_mvc_get_Post($var=NULL) {
    
    // get from post...
    if(is_null($var)) {
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' required a non NULL argument');
    }
    
    if( !empty( $_POST["$var"] ) ) {
        return stripslashes($_POST["$var"]);
    }
    else {
        return FALSE;
    }
    
}

/**
 * Get a value from $_GET
 *
 * @param   string  $var The variable name to retrieve
 * @return  mixed The $_GET var
 * @see     Notes in tina_mvc_get_post()
 */
function tina_mvc_get_Get($var=NULL) {
    
    if(is_null($var)) {
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' required a non NULL argument');
    }
    
    if( !empty( $_GET["$var"] ) ) {
        return stripslashes($_GET["$var"]);
    }
    else {
        return FALSE;
    }
    
}

/**
 * Get a value from $_GET and if no set look in $_POST 
 *
 * @param   string  $var The variable name to retrieve
 * @return  mixed The $_GET/$_POST var
 * @uses    tina_mvc_get_Get() and tina_mvc_get_Post()
 * @see     Notes in tina_mvc_get_post()
 */
function tina_mvc_get_GetPost($var) {
    
    // get from get or post...
    if($ret=_get_Get($var)) {
        return $ret;
    }
    else {
        return _get_Post($var);
    }
    
}

/**
 * Find the path to the Tina MVC plugin folder
 *
 * @return  string The path to the Tina MVC plugin folder (no trailing slash)
 */
function tina_mvc_find_plugin_folder() {
    return (WP_PLUGIN_DIR.'/tina-mvc');
}

/**
 * Find the path to the Tina MVC app folder
 *
 * This is where you should place your own page controllers, views and models
 *
 * @return  string the path to the Tina MVC app folder (no trailing slash)
 * @uses    tina_mvc_find_plugin_folder()
 */
function tina_mvc_find_app_folder() {
    return tina_mvc_find_plugin_folder().'/app';
}

/**
 * Find the path to the Tina MVC libs folder
 *
 * A sensible place for 3rd party libraries (e.g. reCaptcha)
 *
 * @return  string the path to the Tina MVC libs folder (no trailing slash)
 * @uses    tina_mvc_find_plugin_folder()
 */
function tina_mvc_find_libs_folder() {
    return tina_mvc_find_plugin_folder().'/libs';
}

/**
 * Find the path to the Tina MVC emails folder
 *
 * Email templates are placed here
 *
 * @return  string the path to the Tina MVC emails folder (no trailing slash)
 * @uses    tina_mvc_find_plugin_folder()
 * @deprecated since 2010-02-15
 */
function tina_mvc_find_emails_folder() {
    tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' has been depreciated since 2010-02-15.');
    return tina_mvc_find_plugin_folder().'/emails';
}

/**
 * Find the path to the Tina MVC tina_mvc folder
 *
 * All base Tina MVC files live here (controllers, views, etc)
 *
 * @return  string the path to the Tina MVC libs folder (no trailing slash)
 * @uses    tina_mvc_find_plugin_folder()
 * @todo    Perhaps place Tina MVC email templates here too
 */
function tina_mvc_find_tina_mvc_folder() {
    return tina_mvc_find_plugin_folder().'/tina_mvc';
}

/**
 * Make a url to a Tina MVC controller
 *
 * For example, you want to call 'my-app/my-action' and your current front end page controller is 'tina-mvc'
 * this will give you http://example.com/tina-mvc/my-app/my-action. This will fail when used from shortcodes and widgets
 * (because they are not accessed through a front end page controller). In that case you can specify
 * the absolute path to your controller by setting $absolute_controller_path to `true`
 *
 * This function can be called directly from with your controllers (for example if you want to do a browser header redirect)
 * but is normally used in view files (templates) with the tina_mvc_make_controller_link()
 * and tina_mvc_make_abs_controller_link() functions
 *
 * @param   string $controller The 'controller/action/data' we want to call
 * @param   boolean $absolute_controller_path set to `true` to prevent pre-pending the current front end controller to the url (for use in shortcodes and widgets). In this case you must specify the front end page controller in the contoller path e.g. 'tina-mvc/my-page-controller/some-action'
 *
 * @see     tina_mvc_make_controller_link() and tina_mvc_make_abs_controller_link()
 * @return  string An absolute URL to the controller
 * @todo   perhaps get a default front end controller page to direct someone to instead of get_option('home')...
 */
function tina_mvc_make_controller_url($controller='', $absolute_controller_path=false) {
    
    if( $absolute_controller_path ) {
          $_page_link = get_option('home') . "/$controller";
    }
    else {
      if( defined('TINA_MVC_PAGE_CONTROLLER_ID') ) { // no permalinks
          $_page_link = get_page_link( TINA_MVC_PAGE_CONTROLLER_ID ) . "/$controller";
      }
      elseif( defined('TINA_MVC_PAGE_CONTROLLER_NAME') ) {
          $_page_link = get_option('home') . '/' . TINA_MVC_PAGE_CONTROLLER_NAME . "/$controller";
      }
      else {
        
        $extraErrorMessage = "<pre><small>".print_r(debug_backtrace(),1)."</small></pre>";
        
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.': call to controller \''.$controller.'\' failed. `$absolute_controller_path==false` can ony be used when called from a front end controller page (i.e. NOT from a Widget or shortcode and NOT from a hook or action).'.$extraErrorMessage);
      }
    }
    
    return  $_page_link;
}

/**
 * Make a HTML link a Tina MVC controller
 *
 * Uses the current active Tina MVC front end page controller.
 *
 * @param   string $controller The 'controller/action/data' we want to call
 * @param   string $link_text The link text to display in the <a> element (default $controller)
 * @param   string $extra_attribs Attributes to put in the <a> tag. e.g. style or script attributes
 * @return  string An <a> element ready for browser output
 * @uses    tina_mvc_make_controller_url()
 */
function tina_mvc_make_controller_link($controller='', $link_text=FALSE, $extra_attribs='') {
    
    $ret = '<a href="' . tina_mvc_make_controller_url($controller) . '"';
    $ret .=  (  $extra_attribs  ?  ' ' . $extra_attribs  : '' );
    $ret .= '>' . (  $link_text  ?  $link_text  :  $controller )  . '</a>';
    return  $ret;
    
}

/**
 * Make an HTML link a Tina MVC controller.
 *
 * Uses this to make links to arbitrary Tina MVC front end page controllers. You should use this function in
 * your widget and shortcode page controllers when there is no active Tina MVC front end controller.
 * 
 * @param   string $controller The 'controller/action/data' we want to call
 * @param   string $link_text The link text to display in the <a> element (default $controller)
 * @param   string $extra_attribs Attributes to put in the <a> tag. e.g. style or script attributes
 * @return  string An <a> element ready for browser output
 * @uses    tina_mvc_make_controller_url()
 */
function tina_mvc_make_abs_controller_link($controller='', $link_text=FALSE, $extra_attribs='') {
    
    $ret = '<a href="' . tina_mvc_make_controller_url($controller, true) . '"';
    $ret .=  (  $extra_attribs  ?  ' ' . $extra_attribs  : '' );
    $ret .= '>' . (  $link_text  ?  $link_text  :  $controller )  . '</a>';
    return  $ret;
    
}

/**
 * Make a generic 'From:' email address
 *
 * Used by tina_mvc_mail() to make sure there is a consistent 'From:' address for any
 * emails generated by Tina MVC
 *
 * @param   string $local_user The local username to use. (Default: 'no-reply')
 * @return  string USER@SERVER email address
 * @uses    $_SERVER['SERVER_NAME']
 */
function tina_mvc_make_mail_from_address( $local_user='no-reply' ) {
    return 'no_reply@'.$_SERVER['SERVER_NAME'];
}

/**
 * Send an email
 *
 * Grabs an email message template and merges it with any variables you pass. Use this for
 * any emails you want to send through Tina MVC
 * 
 * @param   string $to The recipients address
 * @param   string $message_template Template to use (looks in 'emails' folder)
 * @param   mixed $message_variables Data to be merged into the message (usually array or object )
 * @uses    tina_mvc_make_mail_from_address()
 * @todo    sending attachments
 */
function tina_mvc_mail($to, $message_template=FALSE, $message_variables=FALSE ) {
    
    if( !$to OR !$message_template ) {
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' requires $to and $message_template arguments');
    }
    
    if( ! is_email($to) ) {
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' "'.$to.'" is not a valid email address');
    }
    
    /*
    if( ! array_key_exists( 'subject', $message_variables) OR ! array_key_exists( 'body', $message_variables) ) {
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' array() $message_variables must keys \'subject\' and \'body\'.');
    }
    */
    
    $headers = 'From: '.get_bloginfo( 'name' ).' Mailer <'.tina_mvc_make_mail_from_address().'>' . "\r\n\\";
    
    $tpl = tina_mvc_find_plugin_folder().'/app_emails/'.$message_template.'.php';
    
    if( ! file_exists($tpl) ) {
      $tpl = tina_mvc_find_plugin_folder().'/tina_mvc_emails/'.$message_template.'.php';
    }
    
    if( ! file_exists($tpl) ) {
        tina_mvc_error(__FILE__.' :: '.__FUNCTION__.' $message_template "'.$tpl.'" not found in /app_emails or in /tina_mvc_emails.');
    }
    
    include($tpl); // sets $e_subject and $e_body
    
    wp_mail( $to , $e_subject , $e_body, $headers, $attachments=FALSE);
    
}

/**
 * Check if a user has been assigned to a role
 *
 * This is safe to use when users can have multiple roles
 * 
 * @param   array $roles_to_check A list roles to check
 * @return  bolean
 */
function tina_mvc_user_has_role( $roles_to_check=array() ) {
    
    if( ! $roles_to_check ) return FALSE;
    
    global $current_user;
    get_currentuserinfo();
    $user_id = intval( $current_user->ID );

    if( ! $user_id ) {
      return FALSE;
    }
    
    $user = new WP_User( $user_id ); // $user->roles
    
    foreach( $roles_to_check AS $r) {
      if( in_array($r, $user->roles, FALSE) ) {
        return TRUE;
      }
    }
    
    return FALSE;

}

/**
 * Check if a user has been assigned a capability
 *
 * @param   array $cap_to_check A list capabilities to check
 * @return  bolean
 */
function tina_mvc_user_has_capability( $cap_to_check=array() ) {
    
    if( ! $cap_to_check ) return FALSE;

    global $current_user;
    get_currentuserinfo();
    $user_id = intval( $current_user->ID );
    
    if( ! $user_id ) {
        return FALSE;
    }
    else {
        
        $user = new WP_User( $user_id );
        
        foreach( $cap_to_check AS $c ) {
            if( current_user_can($c) ) return TRUE;
        }
        
    }
    
    return FALSE;  

}

/**
 * Escape a data structure for rendering in a browser
 *
 * Recurses into arrays and objects
 * @param   mixed $data An array or object containing data to be escaped
 * @return  mixed The $escaped $data
 * @uses    ent2ncr() to escape non-XML entities
 */
function tina_mvc_esc_html_recursive( $data=FALSE ) {
    
    if( ! $data ) return FALSE;
    
    if( is_array($data) OR is_object($data) ) {
        
        foreach( $data AS $key => & $value ) {
            
            $key = htmlentities($key,ENT_QUOTES);
            
            if( is_array($value) OR is_object($value) ) {
                tina_mvc_esc_html_recursive( $value );
            }
            elseif( is_string($data) ) {
                $data = ent2ncr(htmlentities($data,ENT_QUOTES));
            }
        }
        
    }
    else {
        $data = htmlentities($data,ENT_QUOTES);
    }
    
    return $data;

}

/**
 * Generate an empty Wordpress 'page' post
 *
 * When a Tina MVC sub controller is called (e.g. `tina/sub-controller`) the
 * function tina_mvc_page_filter() will be passed an empty $posts array. This
 * function creates a default $post
 * 
 * @return  array a Wordpress page
 */
function tina_mvc_get_empty_page_post() {
    
  $p = array(
    'menu_order' => 0, //If new post is a page, sets the order should it appear in the tabs.
    'comment_status' => 'closed', // 'closed' means no comments.
    'comment_count' => 0,
    'ping_status' => 'closed', // 'closed' means pingbacks or trackbacks turned off
    // 'pinged' => [ ? ] //?
    'post_author' => 1, //The user ID number of the author.
    'post_category' => array(1), //Add some categories.
    'post_content' => '', //The full text of the post.
    'post_parent' => '',
    'post_date' => date('Y-m-d H:i:s'), //The time post was made.
    'post_date_gmt' => gmdate('Y-m-d H:i:s'), //The time post was made, in GMT.
    'post_excerpt' => '', //For all your post excerpt needs.
    'post_status' => 'publish', //Set the status of the new post. 
    'post_title' => '', //The title of your post.
    'post_type' => 'page', //Sometimes you want to post a page.
  );
  
  return (object) $p;

}

/**
 * A basic error handler
 *
 * Tina calls this whenever it encounters an error. Hack this to help you
 * debug your applications
 * 
 * @param  string The error message
 * @todo    allow users to override without having to alter a core Tina file
 * @todo    nested HTML...? Mmmmmmm, not really MVC is it?
 */
function tina_mvc_error( $msg ) {
    
    $backtrace = debug_backtrace();
    $base_folder = ABSPATH;
    
    $error  = "<h2>Tina MVC Error</h2>\r\n";
    $error .= "<p><strong>".tina_mvc_esc_html_recursive($msg)."</strong></p>\r\n";
    $error .= "<p><strong>Backtrace:</strong><br><em>NB: file paths are relative to '".tina_mvc_esc_html_recursive($base_folder)."'</em></p>";
    
    $bt_out  = '';
    
    foreach( $backtrace AS $i => & $b ) {
        
        // tiwen at rpgame dot de comment in http://ie2.php.net/manual/en/function.debug-backtrace.php#65433
        if (!isset($b['file'])) $b['file'] = '[PHP Kernel]';
        if (!isset($b['line'])) {
            $b['line'] = 'n/a';   
        }
        else {
            $b['line'] = vsprintf('%s',$b['line']);
        }
        
        $b['function'] = tina_mvc_esc_html_recursive( $b['function'] );
        $b['class'] = tina_mvc_esc_html_recursive( $b['class'] );
        $b['object'] = tina_mvc_esc_html_recursive( $b['object'] );
        $b['type'] = tina_mvc_esc_html_recursive( $b['type'] );
        $b['file'] = tina_mvc_esc_html_recursive(str_replace( $base_folder, '', $b['file']));
        
        if( !empty($b['args']) ) {
            $args = '';
            foreach ($b['args'] as $j => $a) {
                if (!empty($args)) {
                    $args .= "<br>";
                }
                $args .= ' - Arg['.vsprintf('%s',$j).']: ('.gettype($a) . ') '
                      .'<span style="white-space: pre">'.tina_mvc_esc_html_recursive(print_r($a,1)).'</span>';
            }
            
            $b['args'] = $args;
            
        }
        
        $bt_out .= '<strong>['.vsprintf('%s',$i).']: '.$b['file'].' ('.$b['line'].'):</strong><br>';
        $bt_out .= ' - Function: '.$b['function'].'<br>';
        $bt_out .= ' - Class: '.$b['class'].'<br>';
        $bt_out .= ' - Type: '.print_r($b['type'],1).'<br>';
        $bt_out .= ' - Object: '.print_r($b['type'],1).'<br>';
        $bt_out .= $b['args'].'<hr>';
        $bt_out .= "\r\n";
        
    }

    // $error .= "<pre><small>".tina_mvc_esc_html_recursive(print_r($backtrace,1))."</small></pre>\r\n";
    $error .= '<div style="font-size: 70%;">'.$bt_out."</div>\r\n";
              
    wp_die( $error );
    exit();
}

?>
